import sys, struct

# see https://github.com/JoNil/elf2uf2-rs/blob/master/src/uf2.rs
UF2_MAGIC_START0 = 0x0A324655
UF2_MAGIC_START1 = 0x9E5D5157
UF2_MAGIC_END = 0x0AB16F30

UF2_FLAG_NOT_MAIN_FLASH = 0x00000001
UF2_FLAG_FILE_CONTAINER = 0x00001000
UF2_FLAG_FAMILY_ID_PRESENT = 0x00002000
UF2_FLAG_MD5_PRESENT = 0x00004000

RP2040_FAMILY_ID = 0xe48bff56

class UF2Block(object):
    __slots__ = ("bytes", "offset")

    def __init__(self, bytes, offset=0):
        self.bytes = bytes
        self.offset = offset

    @property
    def magic_start0(self):
        return struct.unpack_from("<I", self.bytes, 0)[0]
    @property
    def magic_start1(self):
        return struct.unpack_from("<I", self.bytes, 4)[0]
    @property
    def flags_int(self):
        return struct.unpack_from("<I", self.bytes, 8)[0]
    @property
    def target_addr(self):
        return struct.unpack_from("<I", self.bytes, 12)[0]
    @property
    def payload_size(self):
        return struct.unpack_from("<I", self.bytes, 16)[0]
    @property
    def block_no(self):
        return struct.unpack_from("<I", self.bytes, 20)[0]
    @property
    def num_blocks(self):
        return struct.unpack_from("<I", self.bytes, 24)[0]
    @property
    def file_size_or_family(self):
        return struct.unpack_from("<I", self.bytes, 28)[0]
    @property
    def magic_end(self):
        return struct.unpack_from("<I", self.bytes, 512-4)[0]
    @property
    def data(self):
        return self.bytes[32:512-4]
    @property
    def payload(self):
        return self.data[0..self.payload_size]

    @property
    def is_uf2(self):
        return self.magic_start0 == UF2_MAGIC_START0 and self.magic_start1 == UF2_MAGIC_START1 and self.magic_end == UF2_MAGIC_END

    @property
    def not_valid_reason(self):
        if not self.is_uf2:
            return "no UF2 magic"
        if self.block_no >= self.num_blocks or self.num_blocks == 0:
            return "invalid block number"
        if (self.flags_int & ~(UF2_FLAG_NOT_MAIN_FLASH | UF2_FLAG_FILE_CONTAINER | UF2_FLAG_FAMILY_ID_PRESENT | UF2_FLAG_MD5_PRESENT)) != 0:
            return "unsupported flags"
        return True
    @property
    def valid(self):
        return self.not_valid_reason == True

    @property
    def flags_str(self):
        flags_str = []
        flags2 = self.flags_int
        if (flags2 & UF2_FLAG_NOT_MAIN_FLASH) != 0:
            flags2 &= ~UF2_FLAG_NOT_MAIN_FLASH
            flags_str.append("not_main")
        if (flags2 & UF2_FLAG_FILE_CONTAINER) != 0:
            flags2 &= ~UF2_FLAG_FILE_CONTAINER
            flags_str.append("file")
        if (flags2 & UF2_FLAG_FAMILY_ID_PRESENT) != 0:
            flags2 &= ~UF2_FLAG_FAMILY_ID_PRESENT
            flags_str.append("family")
        if (flags2 & UF2_FLAG_MD5_PRESENT) != 0:
            flags2 &= ~UF2_FLAG_MD5_PRESENT
            flags_str.append("md5")
        if flags2 != 0:
            flags_str.append("0x08x" % flags2)
        return tuple(flags_str)

    @property
    def family(self):
        if (self.flags_int & UF2_FLAG_FAMILY_ID_PRESENT) != 0:
            if self.file_size_or_family == RP2040_FAMILY_ID:
                return "rp2040"
            else:
                return "0x%08x" % self.file_size_or_family
        else:
                return "none"

def read_blocks(filename):
    with open(filename, "rb") as f:
        contents = f.read()

    blocks = []
    offset = 0
    while offset < len(contents):
        block = contents[offset:offset+512]
        block = UF2Block(block, offset=offset)
        if not block.valid:
            raise Exception("no valid UF2 at offset %r: %r" % (offset, block.not_valid_reason))
        blocks.append(block)
        offset += 512
    return blocks

if __name__ == "__main__":
    for block in read_blocks(sys.argv[1]):
        print("block %-6s: target 0x%08x, size %3s, block %3s/%3s, %s, %s" % (
            block.offset, block.target_addr, block.payload_size, block.block_no, block.num_blocks, block.family, " ".join(block.flags_str)))
